open Format
open Config

(* command line option *)

let short_usage = "Usage: completion <options> <file>\n"
let filenames = ref []

let c = ref { vm = false; sn_mthd = LPO }
       
(* parsing *)

let print_trs rs =
  printf "VARIABLES: %a@.EQUATIONS:@.%a@.@." 
    (Formatx.print_list pp_print_string " ") (Rules.variables rs)
     Rules.print rs


(* main *)
let options = Arg.align 
  [("-v", Arg.Unit (Config.set_vm c), 
    " verbose mode");
   ("-s", Arg.String (Config.set_sn_enc c), 
    "<x> set termination encoding (x ::= lpo | rpo | kbo)")]

let () =
  Arg.parse options 
   (fun x -> filenames := !filenames @ [x]) short_usage;
  match !filenames with
  | [f] ->  
      let es = Rules.rpl_spcl_char (Read.read_trs f) in
      let timer = Timer.start () in 
      Rules.print_es_std es;
      let trs = Maxcomp.maxcomp es !c in
      printf "COMPLETE TRS:@.RULES:@.%a@.@.SUCCESS@." Rules.print trs;
      Timer.stop timer;
      let secs = Timer.length ~res:Timer.Seconds timer in
      printf "%s %.2f %s@." "Search time:" secs "seconds" 
  | _ -> eprintf "%s%!" short_usage; exit 1
